#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Support for the Ohara Glass catalog

.. codeauthor: Michael J. Hayford
"""
import logging
from .util import Singleton

import numpy as np

from . import glass


class OharaCatalog(glass.GlassCatalogXLSX, metaclass=Singleton):
    #    data_header = 1
    #    data_start = 2
    #    num_glasses = 134
    #    name_col_offset = 1
    #    coef_col_offset = 60
    #    index_col_offset = 4
    nline_str = {'t': 'nt',
                 's': 'ns',
                 'r': 'nr',
                 'C': 'nC',
                 "C'": "nC'",
                 'D': 'nD',
                 'd': 'nd',
                 'e': 'ne',
                 'F': 'nF',
                 "F'": "nF'",
                 'g': 'ng',
                 'h': 'nh',
                 'i': 'ni'}

    def __init__(self, fname='OHARA.xlsx'):
        super().__init__('Ohara', fname, 'Glass ', 'A1', 'n2325',
                         transmission_offset=81, num_wvls=32)

    def create_glass(self, gname, gcat):
        return OharaGlass(gname)


class OharaGlass(glass.Glass):
    catalog = OharaCatalog()

    def __init__(self, gname, catalog=None):
        if catalog is not None:
            self.catalog = catalog
        super().__init__(gname)

    def glass_code(self):
        return super().glass_code('nd', 'νd')

    def calc_rindex(self, wv_nm):
        wv = 0.001*wv_nm
        wv2 = wv*wv
        coefs = self.coefs
        n2 = 1 + coefs[0]*wv2/(wv2 - coefs[3])
        n2 += coefs[1]*wv2/(wv2 - coefs[4])
        n2 += coefs[2]*wv2/(wv2 - coefs[5])
        return np.sqrt(n2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
"""Glass map display, via Matplotlib

.. Created on Thu Sep 24 22:09:48 2020

.. codeauthor: Michael J. Hayford
"""
import logging
import numpy as np

from matplotlib.figure import Figure
from matplotlib.patches import Polygon
from matplotlib.transforms import Bbox

from . import glass as cat_glass
from . import glassfactory as gf
from . import glasspolygons as gp
from . import util


class GlassMapDB():
    """ Simple model to support Model/View architecture for Glass map views

    Attributes:
        catalogs: list of objects that respond to :meth:`glass_map_data`
    """

    def __init__(self, *args):
        """Initialize a GlassMapDb from a list of args

        Args:
            args: list of items to be included in the GlassMapDB:

                - a dict of catalog names and their catalog instance
                - a list of catalog names
                - a list of :class:`~.glass.Glass` instances

        If no arguments, use the default set of catalogs in
        :mod:`~.glassfactory`, _catalog_list.
        """
        if len(args) == 0:
            args = (gf._catalog_list,)

        self.catalogs = []
        for arg in args:
            if isinstance(arg, dict):
                for cat_name, catalog in arg.items():
                    self.catalogs.append((catalog, cat_name))
            if isinstance(arg, list):
                if isinstance(arg[0], str):
                    # treat as a list of catalog names
                    for cat_name in arg:
                        catalog = gf.get_glass_catalog(cat_name)
                        self.catalogs.append((catalog, cat_name))
                else:
                    # treat as glass instances, wrap them in a GlassMapSet
                    cat_name = 'System'
                    catalog = GlassMapSet(arg, cat_name)
                    self.catalogs.append((catalog, cat_name))

    def get_data_at(self, i, **kwargs):
        catalog, cat_name = self.catalogs[i]
        return catalog.glass_map_data(cat_name=cat_name, **kwargs)

    def get_data_set_label_at(self, i):
        return self.catalogs[i][1]


class GlassMapSet():
    """Set of glass instances to be displayed in a GlassMapFigure. """

    def __init__(self, glasses, cat_name):
        """The main requirement is the glass instance must respond to
        rindex() api
        """
        self.glasses = glasses
        self.cat_name = cat_name

    def catalog_name(self):
        return self.cat_name

    def glass_map_data(self, wvl='d', **kwargs):
        return calc_glass_map_arrays(self.glasses, wvl, 'F', 'C', **kwargs)


def calc_glass_map_arrays(glasses, d_str, F_str, C_str, **kwargs):
    """ return index and dispersion data arrays for input spectral range

    Args:
        nd_str (str): central wavelength string
        nf_str (str): blue end wavelength string
        nc_str (str): red end wavelength string

    Returns:
        index, V-number, partial dispersion, Buchdahl coefficients, and
        glass names
    """

    nd = np.array([g.rindex(d_str) for g in glasses])
    nF = np.array([g.rindex(F_str) for g in glasses])
    nC = np.array([g.rindex(C_str) for g in glasses])

    vd, PCd = cat_glass.calc_glass_constants(nd, nF, nC)

    nd, coefs = cat_glass.calc_buchdahl_coords(
        nd, nF, nC, wlns=(d_str, F_str, C_str), **kwargs)

    names = [g.name()+'/'+g.catalog_name() for g in glasses]
    return nd, vd, PCd, coefs[0], coefs[1], names


class GlassMapFigure(Figure):
    """Matplotlib implementation of an optical glass map.

    Attributes:
        glass_db: an instance of :class:`~.GlassMapDB`
        db_display: list of boolean flags to control catalog display
        hover_glass_names: if True display glass name list under cursor
        plot_display_type: controls the type of data display. Supported types are:

            - "Refractive Index"
            - "Partial Dispersion"
            - "Buchdahl Coefficients"
            - "Buchdahl Dispersion Coefficients"

        refresh_gui: an optional function called when a glass is picked
        pick_list: list of glasses selected by a mouse click. The on_pick fct accumulates the pick_list. Filled with:

                catalog_name, glass_name, nd, vd, PCd

    """
    dsc = [(56/255, 142/255, 142/255),  # sgi teal
           (133/255, 133/255, 133/255),  # grey 52
           (113/255, 113/255, 198/255),  # sgi slateblue
           (102/255, 205/255, 0),  # chartreuse 3
           (255/255, 114/255, 86/255),  # coral 1
           (255/255, 165/255, 0/255),  # orange 1
           (139/255, 139/255, 131/255),  # ivory 4
           ]
    mkr = ['^', 'x', '2', 's', 'v', '+', '*', 'D', 'o']
    home_bbox = Bbox(np.array([[95., 1.45], [20., 2.05]]))

    def __init__(self, glass_db, db_display=None, hover_glass_names=True,
                 plot_display_type="Refractive Index",
                 refresh_gui=None, **kwargs):
        """GlassMap figure initialization. """
        super().__init__(**kwargs)
        self.refresh_gui = refresh_gui
        self.glass_db = glass_db
        num_catalogs = len(glass_db.catalogs)
        self.db_display = db_display if db_display else [True]*num_catalogs
        self.plot_display_type = plot_display_type
        self.hover_glass_names = hover_glass_names
        self.needsClear = True
        self.pick_list = []
        self.event_dict = {}

        self.update_data()

    def connect_events(self, action_dict=None):
        'connect to all the events we need'
        if action_dict is None:
            action_dict = {'motion_notify_event': self.on_hover,
                           # 'button_press_event': self.on_press,
                           'pick_event': self.on_pick,
                           }
        self.callback_ids = []
        for event, action in action_dict.items():
            self.event_dict[event] = action
            cid = self.canvas.mpl_connect(event, action)
            self.callback_ids.append(cid)

    def disconnect_events(self):
        'disconnect all the stored connection ids'
        for clbk in self.callback_ids:
            self.canvas.mpl_disconnect(clbk)
        self.callback_ids = None
        event_dict, self.event_dict = self.event_dict, {}
        return event_dict

    def get_display_label(self):
        return self.plot_display_type

    def refresh(self, **kwargs):
        """Call update_data() followed by plot(), return self.

        Args:
            kwargs: keyword arguments are passed to update_data

        Returns:
            self (class Figure) so scripting envs will auto display results
        """
        self.update_data(**kwargs)
        self.plot()
        return self

    def update_data(self, **kwargs):
        """Fill in raw_data array.

        The raw_data attribute is a list over catalogs. Each catalog has an
        item consisting of the catalog name and a tuple of vectors:

            n, v, p, coefs0, coefs1, glass_names

        """
        self.rawData = []
        ctyp = ("disp_coefs"
                if self.plot_display_type == "Buchdahl Dispersion Coefficients"
                else None)
        for i, display in enumerate(self.db_display):
            gmap_data = self.glass_db.get_data_at(i, ctype=ctyp)
            n, v, p, coefs0, coefs1, glass_names = gmap_data
            catalog_name = self.glass_db.get_data_set_label_at(i)
            self.rawData.append([catalog_name,
                                 (n, v, p, coefs0, coefs1, glass_names)])
        return self

    def update_axis_limits(self, bbox):
        self.ax.set_xlim(bbox[0][0], bbox[1][0])
        self.ax.set_ylim(bbox[0][1], bbox[1][1])

    def draw_axes(self):
        self.ax.grid(True)
        if hasattr(self, 'header'):
            self.ax.set_title(self.header, pad=10.0, fontsize=18)
        if hasattr(self, 'x_label'):
            self.ax.set_xlabel(self.x_label)
        if hasattr(self, 'y_label'):
            self.ax.set_ylabel(self.y_label)

    def plot(self):
        try:
            self.ax.cla()
        except AttributeError:
            self.ax = self.add_subplot(1, 1, 1)

        if self.plot_display_type == "Refractive Index":
            self.x_label = r'$\mathrm{V_d}$'
            self.y_label = r'$\mathrm{n_d}$'
            xi = 1
            yi = 0
            self.draw_glass_polygons()
        elif self.plot_display_type == "Partial Dispersion":
            self.x_label = r'$\mathrm{V_d}$'
            self.y_label = r'$\mathrm{P_{F-d}}$'
            xi = 1
            yi = 2
        elif self.plot_display_type == "Buchdahl Coefficients":
            self.x_label = r'$\mathrm{\nu_2}$'
            self.y_label = r'$\mathrm{\nu_1}$'
            xi = 4
            yi = 3
        elif self.plot_display_type == "Buchdahl Dispersion Coefficients":
            self.x_label = r'$\mathrm{\eta_2}$'
            self.y_label = r'$\mathrm{\eta_1}$'
            xi = 4
            yi = 3
        self.ax.set_title(self.get_display_label())
        for i, display in enumerate(self.db_display):
            line = self.ax.plot(self.rawData[i][1][xi], self.rawData[i][1][yi],
                                linestyle='None', marker='o', markersize=5,
                                # linestyle='None', markersize=7,
                                alpha=0.75, gid=i,
                                picker=True, pickradius=5,
                                color=self.dsc[i],
                                # marker=self.mkr[i], fillstyle='none',
                                label=self.rawData[i][0], visible=display)
            # set pickradius here because of a bug. Fixed in 3.3
            line[0].set_pickradius(5.)

        if self.plot_display_type == "Refractive Index":
            # provide a default minimum area, and update view limits
            # accordingly
            viewLim = Bbox.union([self.home_bbox, self.ax.viewLim])
            self.update_axis_limits(viewLim.get_points())

        # set up interactive event handling
        # The pick events, one per artist, are sent before the sole button
        # press event
        actions = {'button_press_event': self.on_press,
                   'pick_event': self.on_pick,
                   }
        if self.hover_glass_names:
            actions['motion_notify_event'] = self.on_hover

        self.connect_events(action_dict=actions)

        # set up hover annotation
        if self.hover_glass_names:
            self.hover_list = self.ax.annotate(
                "", xy=(0, 0), xytext=(20, 20),
                textcoords="offset points",
                bbox=dict(boxstyle="round", fc="w"),
                arrowprops=dict(arrowstyle="->"))
            self.hover_list.set_visible(False)

        # draw remaining stuff, axes, legend...
        if xi == 1:
            self.ax.invert_xaxis()
        self.draw_axes()
        self.ax.legend()
        self.canvas.draw()
        return self

    def draw_glass_polygons(self):
        for glass, poly in gp.polygons.items():
            rgb = gp.rgb[glass]
            p = Polygon(poly, closed=True, fc=util.rgb2mpl(rgb), ec='black',
                        linewidth=1.0)
            self.ax.add_artist(p)

    def clear_pick_table(self):
        self.pick_list = []
        self.needsClear = False

    # --- interactive actions
    def find_artists_at_location(self, event):
        """Returns a list of shapes in zorder at the event location."""
        artists = []
        for artist in self.ax.get_children():
            inside, info = artist.contains(event)
            if inside:
                id = artist.get_gid()
                if id is not None:
                    artists.append((artist, info, id))

        return sorted(artists, key=lambda a: a[0].get_zorder(),
                      reverse=True)

    def on_hover(self, event):
        vis = self.hover_list.get_visible()
        artists = self.find_artists_at_location(event)
        info_text = []
        if len(artists) > 0:
            for a in artists:
                artist, info, cat = a
                if self.db_display[cat]:
                    ind = info['ind']
                    cat_name = self.rawData[cat][0]
                    n, v, p, coef0, coef1, glass_name = self.rawData[cat][1]
                    for k in ind:
                        text = glass_name[k] + ', ' + cat_name
                        info_text.append(text)
            # Update annotation with glass list
            info_text = '\n'.join(info_text)
            self.hover_list.set_text(info_text)
            pos = event.xdata, event.ydata
            self.hover_list.xy = pos
            self.hover_list.get_bbox_patch().set_alpha(0.8)
            self.hover_list.set_visible(True)
            self.canvas.draw_idle()
        else:
            if vis:
                self.hover_list.set_visible(False)
                self.canvas.draw_idle()

    def on_pick(self, event):
        """ handle picking glasses under the cursor.

        One pick event for each catalog, extract selected glasses and add to
        pick_list
        """
        logging.debug("on_pick: needsClear={}".format(self.needsClear))
        if self.needsClear:
            self.clear_pick_table()
        line = event.artist
        cat = line.get_gid()
        if self.db_display[cat]:
            ind = event.ind
            cat_name = self.rawData[cat][0]
            n, v, p, coef0, coef1, glass_name = self.rawData[cat][1]
            for k in ind:
                glass = (cat_name, glass_name[k], n[k], v[k], p[k])
                self.pick_list.append(glass)

    def on_press(self, event):
        """ handle mouse clicks within the diagram.

        The button press event is sent after the pick events; it will be sent
        in cases with no pick events, e.g. clicking in an empty area of the
        axes. The two cases are:

            - if there were pick events, needsClear will be False so that items
              from different artists can be accumulated in the pick_list. The
              press event signals no further item accumulation. Flip needsClear
              to True so the next pick or press event will clear the pick_list.

            - if there were no pick events, needsClear will be True. Call
              clear_pick_table to empty pick_list and reset needsClear to False.

        """
        logging.debug("on_press: needsClear={}".format(self.needsClear))
        if self.needsClear:
            # If needsClear is still set, there have been no pick events so
            #  this is a click in an empty region of the plot.
            #  Clear the pick table
            self.clear_pick_table()
        else:
            # on_press event happens after on_pick events. Set needsClear for
            #  next on_pick, i.e. a new selection, to handle
            self.needsClear = True
        if self.refresh_gui is not None:
            self.refresh_gui()

    def updateVisibility(self, indx, state):
        self.ax.lines[indx].set_visible(state)
        self.canvas.draw()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
""" Support for Glass catalogs and instances

The ``glass`` module contains the two base classes fundamental to the
:mod:`opticalglass` module. The :class:`~opticalglass.glass.GlassCatalog` class
implements as much of the common functionality needed for access to the
catalog data as possible.

The :class:`~opticalglass.glass.Glass` is an interface to the data for a particular
glass in a catalog. The primary function of interest for optical calculations
is :func:`~opticalglass.glass.Glass.rindex` which returns the refractive index at the
input wavelength (nm).

A factory interface to ``Glass`` creation is the function
:func:`~opticalglass.glassfactory.create_glass` that returns a ``Glass``
instance of the appropriate catalog type, given the glass and catalog names.

.. codeauthor: Michael J. Hayford
"""
import itertools
import logging
import warnings
from pathlib import Path

import xlrd
from openpyxl import load_workbook

import numpy as np
from scipy import linalg

from . import buchdahl
from . import glasserror as ge
from .util import Singleton, Counter
from .spectral_lines import get_wavelength

from .caselessDictionary import CaselessDictionary


def get_filepath(fname):
    """ given a (spreadsheet) file name, return a complete Path to the file

    The data files included with the ``opticalglass`` package are located in
    a data directory in the package hierarchy.
    ::

        opticalglass/
            opticalglass/
                data/
                    fname

    Args:
        fname (str): the spreadsheet filename, including extender

    Returns:
        str: full path including filename for requested spreadsheet
    """
    pth = Path(__file__).resolve().parent
    return pth/'data'/fname


def calc_glass_constants(nd, nF, nC):
    """Given central, blue and red refractive indices, calculate Vd and PFd."""
    dFC = nF-nC
    vd = (nd - 1.0)/dFC
    PFd = (nF-nd)/dFC
    return vd, PFd


def calc_buchdahl_coords(nd, nF, nC, wlns=('d', 'F', 'C'),
                         ctype=None, **kwargs):
    """Given central, blue and red refractive indices, calculate the Buchdahl
    chromatic coefficients.

    Arguments:
        nd: central refractive index
        nF: "blue" refractive index
        nC: "red" refractive index
        wlns: wavelengths for the 3 refractive indices
        ctype: if "disp_coefs", return dispersion coefficients, otherwise the
               quadratic coefficients
    """
    wv0 = buchdahl.get_wv(wlns[0])

    omF = buchdahl.omega(buchdahl.get_wv(wlns[1]) - wv0)
    omC = buchdahl.omega(buchdahl.get_wv(wlns[2]) - wv0)

    a = np.array([[omF, omF**2], [omC, omC**2]])
    b = np.array([nF-nd, nC-nd])
    coefs = np.linalg.solve(a, b)
    if ctype == "disp_coefs":
        coefs /= (nd - 1)
    return nd, coefs


def fit_buchdahl_coords(indices, degree=2,
                        wlns=['d', 'h', 'g', 'F', 'e', 'C', 'r']):
    """Given central, 4 blue and 2 red refractive indices, do a least squares
    fit for the Buchdahl chromatic coefficients.
    """
    rind0 = indices[0]
    wv0 = buchdahl.get_wv(wlns[0])
    om = [buchdahl.omega(buchdahl.get_wv(w) - wv0) for w in wlns]

    a = np.array([[o**(i+1) for i in range(degree)] for o in om])
    b = np.array(indices) - rind0

    results = linalg.lstsq(a, b)
    coefs = results[0]

    return rind0, coefs


def get_glass_map_arrays(cat, d_str, F_str, C_str, **kwargs):
    """ return index and dispersion data arrays for input spectral range

    Args:
        nd_str (str): central wavelength string
        nf_str (str): blue end wavelength string
        nc_str (str): red end wavelength string

    Returns:
        index, V-number, partial dispersion, Buchdahl coefficients, and
        glass names
    """
    nd = np.array(
            cat.catalog_data(cat.data_index(cat.nline_str[d_str])))
    nF = np.array(
            cat.catalog_data(cat.data_index(cat.nline_str[F_str])))
    nC = np.array(
            cat.catalog_data(cat.data_index(cat.nline_str[C_str])))

    vd, PFd = calc_glass_constants(nd, nF, nC)

    nd, coefs = calc_buchdahl_coords(
        nd, nF, nC, wlns=(d_str, F_str, C_str), **kwargs)

    names = cat.get_glass_names()
    return nd, vd, PFd, coefs[0], coefs[1], names


class GlassCatalogSpreadsheet:
    """ spreadsheet-based glass catalog """

    def get_glass_names(self):
        """ returns a list of glass names """
        pass

    def get_column_names(self, dindex=None, num=None, add_offset=0):
        """ returns a list of column headers """
        pass

    def get_data_for_glass(self, gindex, dindex=None, num=None):
        """ returns an array of glass data for the glass at *gindex*

        Args:
            gindex: glass index into spreadsheet
            dindex: the starting column of the desired data
            num: number of data items (cells) to retrieve

        Returns: list of data items
        """
        pass

    def get_transmission_wvl(self, header_str):
        """Returns the wavelength value from the transmission data header string."""
        pass

    def transmission_data(self, gindex):
        """ returns an array of transmission data for the glass at *gindex*

        Args:
            gindex: glass index into spreadsheet
            transmission_offset: the starting column of the transmission data
            num_wvls: number of wavelengths for the data

        Returns:
            list of wavelength, transmittance pairs
        """
        header = self.get_column_names(
            dindex=self.transmission_offset,
            num=self.num_wvls,
            add_offset=self.transmission_header_offset)

        data = self.get_data_for_glass(gindex,
                                       dindex=self.transmission_offset,
                                       num=self.num_wvls)

        trns_data = []
        for h, d in zip(header, data):
            w = self.get_transmission_wvl(h)
            if d is not None and d != '' and d != ' ' and d != '-':
                trns_data.append((w, d))
        return trns_data


class GlassCatalogXLS(GlassCatalogSpreadsheet):
    """ Older, Excel XLS-based glass catalog

    Args:
        name: name of the glass catalog
        fname: excel filename, located in ``data`` directory
        glass_str: the header string for the Glass column in fname
        coef_str: the header string for the first refractive index coefficient
                  column in fname
        rindex_str: the header string for the first refractive index value
                    column in fname

    Attributes:
        num_glasses: number of glasses in the catalog
        num_columns: number of colums in the spreadsheet
        data_header: the row containing the data header labels
        data_start: first row in the spreadsheet contain glass data
        name_col_offset: the column offset for glass_str
        coef_col_offset: the column offset for coef_str
        num_coefs: the number of coefficients in the index equation
        index_col_offset: the column offset for rindex_str
        data_header_offset: the row offset of the data headers from the
                            glass_str row
        nline_str: a dict of strings mapping header strings used for index for
                   spectral lines. **Must be provided by the derived class**
    """

    def __init__(self, name, fname, glass_str, coef_str, rindex_str,
                 num_coefs=6, data_header_offset=0, glass_name_offset=1,
                 transmission_offset=0, num_wvls=0,
                 transmission_header_offset=0):
        # data_index is zero-based
        self.based = 0
        self.name = name
        # Open the workbook
        xl_workbook = xlrd.open_workbook(get_filepath(fname))
        self.xl_data = xl_workbook.sheet_by_index(0)

        for i in range(0, self.xl_data.nrows):
            try:
                self.name_col_offset = (self.xl_data.row_values(i, 0)
                                        .index(glass_str))
                glass_header = i
                self.data_header = i
                # the data headers may be offset from the Glass header row
                self.data_header += data_header_offset
                break
            except ValueError:
                pass

        for j in range(glass_header+glass_name_offset, self.xl_data.nrows):
            gname = self.xl_data.cell_value(j, self.name_col_offset)
            if len(gname) > 0:
                self.data_start = j
                break

        gnames = self.get_glass_names()
        self.num_glasses = len(gnames)

        colnames = self.xl_data.row_values(self.data_header, 0)
        self.coef_col_offset = colnames.index(coef_str)
        self.num_coefs = num_coefs
        self.index_col_offset = colnames.index(rindex_str)

        # data slice for 10mm transmission
        self.transmission_offset = transmission_offset
        self.num_wvls = num_wvls
        self.transmission_header_offset = transmission_header_offset

        # build an alphabetical list of decoded glass names
        glass_list = [(decode_glass_name(gn), gn, name)
                      for gn in self.get_glass_names()]
        glass_list = sorted(glass_list, key=lambda glass: glass[0][0])
        # build a lookup dict of the glass defs keyed to decoded glass names
        glass_lookup = {gn_decode: (gn, gc)
                        for gn_decode, gn, gc in glass_list}
        # attach these 'static' lists to class variables
        self.__class__.glass_list = glass_list
        self.__class__.glass_lookup = glass_lookup

    def catalog_name(self):
        return self.name

    def get_glass_names(self):
        """ returns a list of glass names """
        gnames = self.xl_data.col_values(self.name_col_offset, self.data_start)
        # filter out any empty cells at the end
        while gnames and len(gnames[-1]) == 0:
            gnames.pop()
        return gnames

    def get_column_names(self, dindex=None, num=None, add_offset=0):
        """ returns a list of column headers """
        if dindex is not None:  # get a partial row of at least 1 item
            num = num if num is not None else 1
            col_start = dindex
            row_start = self.data_header + add_offset
            colnames = self.xl_data.row_values(row_start,
                                               col_start, col_start+num)
        else:  # get the entire row
            row_start = self.data_header + add_offset
            colnames = self.xl_data.row_values(row_start, 0)

        return colnames

    def glass_index(self, gname):
        """ returns the glass index (row) for glass name `gname`

        Args:
            gname (str): glass name

        Returns:
            int: the index (row) of the requested glass

        Raises:
            GlassNotFoundError: if **gname** doesn't match any header string
        """
        gnames = self.xl_data.col_values(self.name_col_offset, self.data_start)
        if gname in gnames:
            gindex = gnames.index(gname)
        else:
            logging.info('glass %s not found in catalog %s', gname, self.name)
            raise ge.GlassNotFoundError(self.name, gname)

        return gindex

    def data_index(self, dname):
        """ returns the data index (column) for data `dname`

        Args:
            dname (str): header string for data

        Returns:
            int: the index (column) of the requested data

        Raises:
            GlassDataNotFoundError: if **dname** doesn't match any header string
        """
        if dname in self.xl_data.row_values(self.data_header, 0):
            dindex = self.xl_data.row_values(self.data_header, 0).index(dname)
        else:
            logging.info('data type %s not found in catalog %s', dname, self.name)
            raise ge.GlassDataNotFoundError(self.name, dname)

        return dindex

    def get_data_for_glass(self, gindex, dindex=None, num=None):
        """ returns an array of glass data for the glass at *gindex*

        Args:
            gindex: glass index into spreadsheet
            dindex: the starting column of the desired data
            num: number of data items (cells) to retrieve

        Returns: list of data items
        """
        if dindex is not None:  # get a partial row of at least 1 item
            num = num if num is not None else 1
            col_start = dindex
            glass_data = self.xl_data.row_values(self.data_start+gindex,
                                                 col_start, col_start+num)
        else:  # get the entire row
            glass_data = self.xl_data.row_values(self.data_start+gindex, 0)

        return glass_data

    def glass_data(self, gindex):
        """ returns an array of data for the glass at **gindex** """
        return self.xl_data.row_values(self.data_start+gindex, 0)

    def catalog_data(self, dindex):
        """ returns an array of data at column **dindex** for all glasses """
        return self.xl_data.col_values(dindex, self.data_start,
                                       self.data_start+self.num_glasses)

    def glass_coefs(self, gindex):
        """ returns an array of glass coefficients for the glass at **gindex** """
        return (self.xl_data.row_values(self.data_start+gindex,
                                        self.coef_col_offset,
                                        self.coef_col_offset+self.num_coefs))

    def get_transmission_wvl(self, header_str):
        """Returns the wavelength value from the transmission data header string."""
        return float(header_str)

    def glass_map_data(self, wvl='d', **kwargs):
        """ return index and dispersion data for all glasses in the catalog

        Args:
            wvl (str): the central wavelength for the data, either 'd' or 'e'

        Returns:
            index, V-number, partial dispersion, Buchdahl coefficients, and
            glass names
        """
        return get_glass_map_arrays(self, wvl, 'F', 'C', **kwargs)


class GlassCatalogXLSX(GlassCatalogSpreadsheet):
    """ Excel XLSX-based glass catalog

    Args:
        name: name of the glass catalog
        fname: excel filename, located in ``data`` directory
        glass_str: the header string for the Glass column in fname
        coef_str: the header string for the first refractive index coefficient
                  column in fname
        rindex_str: the header string for the first refractive index value
                    column in fname

    Attributes:
        num_glasses: number of glasses in the catalog
        num_columns: number of colums in the spreadsheet
        data_header: the row containing the data header labels
        data_start: first row in the spreadsheet contain glass data
        name_col_offset: the column offset for glass_str
        coef_col_offset: the column offset for coef_str
        num_coefs: the number of coefficients in the index equation
        index_col_offset: the column offset for rindex_str
        data_header_offset: the row offset of the data headers from the
                            glass_str row
        nline_str: a dict of strings mapping header strings used for index for
                   spectral lines. **Must be provided by the derived class**
    """

    def __init__(self, name, fname, glass_str, coef_str, rindex_str,
                 num_coefs=6, data_header_offset=0, glass_name_offset=1,
                 transmission_offset=0, num_wvls=0,
                 transmission_header_offset=0):
        # data_index is one-based
        self.based = 1
        self.name = name
        # Open the workbook
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=UserWarning)
            wb = load_workbook(get_filepath(fname))
        self.ws = ws = wb.active

        for i, row in enumerate(ws.iter_rows(values_only=True), start=1):
            try:
                name_col_offset = row.index(glass_str)+1
            except ValueError:
                pass
            else:
                data_header = i
                # the data headers may be offset from the Glass header row
                data_header += data_header_offset
                self.data_header = data_header
                self.name_col_offset = name_col_offset
                break

        # find beginning and end of data in names column
        cols = ws.iter_cols(min_col=name_col_offset,
                            max_col=name_col_offset,
                            values_only=True)
        gnames = list(next(cols))
        for j, gname in enumerate(gnames[data_header:], start=data_header+1):
            if gname is not None and len(gname) > 0:
                self.data_start = j
                break

        # filter out any empty cells at the end
        while len(gnames) > 0:
            if gnames[-1] is None:
                gnames.pop()
            elif len(gnames[-1]) == 0:
                gnames.pop()
            else:
                break

        self.num_glasses = len(gnames[self.data_start-1:])

        # find items in column header row
        rows = ws.iter_rows(min_row=data_header,
                            max_row=data_header,
                            values_only=True)

        colnames = list(next(rows))

        # filter out any empty cells at the end
        while len(colnames) > 0:
            if colnames[-1] is None:
                colnames.pop()
            elif len(colnames[-1]) == 0:
                colnames.pop()
            else:
                break

        self.num_columns = len(colnames)

        self.coef_col_offset = colnames.index(coef_str)+1
        self.num_coefs = num_coefs
        self.index_col_offset = colnames.index(rindex_str)+1

        # data slice for 10mm transmission
        self.transmission_offset = transmission_offset
        self.num_wvls = num_wvls
        self.transmission_header_offset = transmission_header_offset

        # build an alphabetical list of decoded glass names
        glass_list = [(decode_glass_name(gn), gn, name)
                      for gn in self.get_glass_names()]
        glass_list = sorted(glass_list, key=lambda glass: glass[0][0])
        # build a lookup dict of the glass defs keyed to decoded glass names
        glass_lookup = {gn_decode: (gn, gc)
                        for gn_decode, gn, gc in glass_list}
        # attach these 'static' lists to class variables
        self.__class__.glass_list = glass_list
        self.__class__.glass_lookup = glass_lookup

    def catalog_name(self):
        return self.name

    def get_glass_names(self):
        """ returns a list of glass names """
        cols = self.ws.iter_cols(min_col=self.name_col_offset,
                                 max_col=self.name_col_offset,
                                 min_row=self.data_start,
                                 max_row=self.data_start+self.num_glasses-1,
                                 values_only=True)
        gnames = list(next(cols))
        return gnames

    def get_column_names(self, dindex=None, num=None, add_offset=0):
        """ returns a list of column headers """
        if dindex is not None:  # get a partial row of at least 1 item
            num = num if num is not None else 1
            col_start = dindex
            row_start = self.data_header + add_offset
            rows = self.ws.iter_rows(min_col=col_start,
                                     max_col=col_start+num-1,
                                     min_row=row_start,
                                     max_row=row_start,
                                     values_only=True)
        else:  # get the entire row
            row_start = self.data_header + add_offset
            rows = self.ws.iter_rows(min_col=1,
                                     max_col=self.num_columns,
                                     min_row=row_start,
                                     max_row=row_start,
                                     values_only=True)

        colnames = list(next(rows))
        return colnames

    def glass_index(self, gname):
        """ returns the glass index (row) for glass name `gname`

        Args:
            gname (str): glass name

        Returns:
            int: the 1-based index (row) of the requested glass

        Raises:
            GlassNotFoundError: if **gname** doesn't match any header string
        """
        gnames = self.get_glass_names()
        if gname in gnames:
            gindex = gnames.index(gname)+1
        else:
            logging.info('glass %s not found in catalog %s', gname, self.name)
            raise ge.GlassNotFoundError(self.name, gname)

        return gindex

    def data_index(self, dname):
        """ returns the data index (column) for data `dname`

        Args:
            dname (str): header string for data

        Returns:
            int: the 1-based index (column) of the requested data

        Raises:
            GlassDataNotFoundError: if *dname* doesn't match any header string
        """
        colnames = self.get_column_names()
        if dname in colnames:
            dindex = colnames.index(dname)+1
        else:
            logging.info('data type %s not found in catalog %s',
                         dname, self.name)
            raise ge.GlassDataNotFoundError(self.name, dname)

        return dindex

    def get_data_for_glass(self, gindex, dindex=None, num=None):
        """ returns an array of glass data for the glass at *gindex*

        Args:
            gindex: glass index into spreadsheet
            dindex: the starting column of the desired data
            num: number of data items (cells) to retrieve

        Returns: list of data items
        """
        if dindex is not None:  # get a partial row of at least 1 item
            num = num if num is not None else 1
            col_start = dindex
            row_start = gindex + self.data_start - 1
            rows = self.ws.iter_rows(min_col=col_start,
                                     max_col=col_start+num-1,
                                     min_row=row_start,
                                     max_row=row_start,
                                     values_only=True)
        else:  # get the entire row
            row_start = gindex + self.data_start - 1
            rows = self.ws.iter_rows(min_col=1,
                                     max_col=self.num_columns,
                                     min_row=row_start,
                                     max_row=row_start,
                                     values_only=True)

        glass_data = next(rows)
        return glass_data

    def glass_data(self, gindex):
        """ returns an array of data for the glass at *gindex* """
        glass_data = self.get_data_for_glass(gindex)
        return glass_data

    def glass_coefs(self, gindex):
        """ returns an array of glass coefficients for the glass at *gindex* """
        glass_data = self.get_data_for_glass(gindex,
                                             dindex=self.coef_col_offset,
                                             num=self.num_coefs)
        return glass_data

    def catalog_data(self, dindex):
        """ returns an array of data at column *dindex* for all glasses """
        cols = self.ws.iter_cols(min_col=dindex,
                                 max_col=dindex,
                                 min_row=self.data_start,
                                 max_row=self.data_start+self.num_glasses-1,
                                 values_only=True)
        return list(next(cols))

    def get_transmission_wvl(self, header_str):
        """Returns the wavelength value from the transmission data header string."""
        return float(header_str)

    def glass_map_data(self, wvl='d', **kwargs):
        """ return index and dispersion data for all glasses in the catalog

        Args:
            wvl (str): the central wavelength for the data, either 'd' or 'e'

        Returns:
            index, V-number, partial dispersion, Buchdahl coefficients, and
            glass names
        """
        return get_glass_map_arrays(self, wvl, 'F', 'C', **kwargs)


class Glass:
    """ base optical glass

    Attributes:
        gname: the glass name
        gindex: the index into the glass list
        catalog: the GlassCatalog this glass is associated with.
                 **Must be provided by the derived class**
        coefs: list of coefficients for calculating refractive index vs wv
    """

    def __init__(self, gname):
        self.gindex = self.catalog.glass_index(gname)
        self.gname = gname
        self.coefs = self.catalog.glass_coefs(self.gindex)

    def __str__(self):
        return self.catalog.name + ' ' + self.name() + ': ' + self.glass_code()

    def __repr__(self):
        return "{!s}('{}')".format(type(self).__name__, self.gname)

    def sync_to_restore(self):
        """ hook routine to restore the gindex given gname """
        self.gindex = self.catalog.glass_index(self.gname)
        if not hasattr(self, 'coefs'):
            self.coefs = self.catalog.glass_coefs(self.gindex)

    def glass_code(self, nd_str, vd_str):
        """ returns the 6 digit glass code, combining index and V-number """
        nd = self.glass_item(nd_str)
        vd = self.glass_item(vd_str)
        return str(1000*round((nd - 1), 3) + round(vd/100, 3))

    def glass_data(self):
        """ returns the raw spreadsheet data for the glass """
        return self.catalog.glass_data(self.gindex)

    def name(self):
        """ returns the glass name, :attr:`gname` """
        return self.gname

    def catalog_name(self):
        """ returns the glass name, :attr:`gname` """
        return self.catalog.catalog_name()

    def glass_item(self, dname):
        """ return the value of the **dname** item

        Args:
            dname (str): header string for data

        Returns:
            the *dname* data

        Raises:
            GlassDataNotFoundError: if *dname* doesn't match any header string
        """
        dindex = self.catalog.data_index(dname)
        if dindex is None:
            return None
        else:
            return self.glass_data()[dindex-self.catalog.based]

    def rindex(self, wvl):
        """ returns the interpolated refractive index at wvl

        Args:
            wvl: either the wavelength in nm or a string with a spectral line
                 identifier. for the refractive index query

        Returns:
            float: the refractive index at wv_nm

        Raises:
            KeyError: if ``wvl`` is not in the spectra dictionary
        """
        return self.calc_rindex(get_wavelength(wvl))

    def calc_rindex(self, wv_nm):
        """ returns the interpolated refractive index at wv_nm

        **Must be provided by the derived class**

        Args:
            wv_nm (float or numpy array): wavelength in nm for the refractive index query

        Returns:
            float or numpy array: the refractive index at wv_nm
        """
        pass

    def transmission_data(self):
        """ returns an array of transmission data for the glass

        Returns: list of wavelength, transmittance pairs for 10mm sample
        """
        return self.catalog.transmission_data(self.gindex)


def decode_glass_name(glass_name):
    """Split glass_name into prefix, group, num, suffix.

    Manufacturers glass names follow a common pattern. At the simplest, it is
    a short character string, typically used to identify a particular glass
    composition, with a numeric qualifier. The composition group and product
    serial number are combined to form the basic product id, the group_num:

        - F2
        - SF56

    Manufacturers will often use a single character prefix to indicate
    different categories of glasses, e.g. moldable or "New":

        - N-BK7
        - P-LASF50

    Similarly, a suffix with one or more characters is often used to
    differentiate between different variations of the same base material.

        - N-SF57
        - N-SF57HT
        - N-SF57HTultra

    This function takes an input glass name and returns a tuple of strings. A
    valid glass_name should always have a non-null group_num; prefixes and
    suffixes are optional and used differently by different manufacturers.

        * group_num, prefix, suffix
        * group, num = group_num

    Args:
        glass_name (str): a glass manufacturer's glass name

    Returns: group_num, prefix, suffix, where group_num = group, num

    """
    gn = glass_name.split('-')
    suffix = ''
    if len(gn) == 1:
        prefix = ''
        gn2 = gn[0]
    elif len(gn) == 3:
        prefix = gn[0]
        suffix = gn[2]
        gn2 = gn[1]
    elif len(gn) == 2:
        if len(gn[0]) < 3:
            prefix = gn[0]
            gn2 = gn[1]
        else:
            prefix = ''
            gn2 = gn[0]
            suffix = gn[1]

    group = gn2
    num = ''
    for i, char in enumerate(gn2):
        if char.isdigit():
            start = i
            while i < len(gn2) and gn2[i].isdigit():
                i += 1
            group = gn2[:start].rstrip()
            num = gn2[start:i]
            break
    suffix = gn2[i:] if suffix == '' else suffix
    group_num = group, num
    return group_num, prefix, suffix


def glass_catalog_stats(glass_list, do_print=False):
    """Decode all of the glass names in glass_cat_name.

    Print out the original glass names and the decoded version side by side.

    Args:
        glass_list: ((group, num), prefix, suffix), glass_name, glass_cat_name
        do_print (bool): if True, print the glass name and the decoded version

    Returns:
        groups (dict): all catalog groups
        group_num (dict): all glasses with multiple variants
        prefixes (dict): all the non-null prefixes used
        suffixes (dict): all the non-null suffixes used
    """
    groups = Counter()
    group_nums = Counter()
    prefixes = Counter()
    suffixes = Counter()
    for g in glass_list:
        (group_num, prefix, suffix), gn, gc = g
        group, num = group_num
        if prefix != '':
            prefixes[prefix] += 1
        if suffix != '':
            suffixes[suffix] += 1
        groups[group] += 1
        group_nums[group_num] += 1
        if do_print:
            fmt_3 = "{:14s} {:>2s}  {:8s}  {:12s}"
            print(fmt_3.format(gn, prefix, group+num, suffix))
    # filter group_nums to include only multi-use items
    group_nums = {k: v for k, v in group_nums.items() if v > 1}
    return groups, group_nums, prefixes, suffixes


class Robb1983Catalog(metaclass=Singleton):
    """ glass catalog based on data in Robb, et als 1983 paper on Buchdahl's
    chromatic coordinate

    The data used by this class is from the 1983 paper by Paul N. Robb and
    R. I. Mercado, `Calculation of refractive indices using Buchdahl’s
    chromatic coordinate <https://doi.org/10.1364/AO.22.001198>`_ . The
    copyright date for all 5 catalogs cited was 1980.

    Args:
        fname: filename, located in ``data`` directory

    Attributes:
        glass_db: dict lookup by catalog and glass name, value is decoded
                  glassname and Buchdahl coefficients
        glass_list: list of decoded_glassname, glassname, and catalog per glass

    """

    _cat_names = ["SCHOTT", "OHARA", "HOYA", "CORNING-FRANCE", "CHANCE"]

    def __init__(self, fname='robb1983_data_final.txt'):
        # Open the workbook
        glass_db = CaselessDictionary()
        glass_list = []
        rndx_list = []
        nu1_list = []
        nu2_list = []
        with get_filepath(fname).open() as f_input:
            for line in f_input:
                if line[0] == '#':
                    if 'GLASS CATALOGUE' in line:
                        tokens = line[1:].split()
                        catalog = 'Robb1983.' + tokens[0]
                        glass_db[catalog] = {}
                else:
                    tokens = line.split()
                    gname = tokens[0]
                    try:
                        gname_decode = decode_glass_name(gname)
                    except UnboundLocalError:
                        print(catalog, gname)
                    else:
                        rndx = float(tokens[1])
                        nu1 = float(tokens[2])
                        nu2 = float(tokens[3])
                        glass_db[catalog][gname] = (
                            gname_decode, rndx, nu1, nu2)
                        glass_list.append((gname_decode, gname, catalog))
                        rndx_list.append(rndx)
                        nu1_list.append(nu1)
                        nu2_list.append(nu2)

        self.glass_db = glass_db
        self.glass_list = glass_list
        self._glass_data = CaselessDictionary()

    @property
    def glass_data(self):
        if len(self._glass_data) == 0:
            for rbk, rbv in self.glass_db.items():
                gnames = list(rbv.keys())
                gdata = np.array([[d[1], d[2], d[3]] for d in rbv.values()])
                self._glass_data[rbk] = gnames, gdata
        return self._glass_data

    def create_glass(self, gname, gcat):
        catalog = gcat if 'Robb1983.' in gcat else 'Robb1983.' + gcat
        try:
            gdata = self.glass_db[catalog][gname]
        except KeyError:
            return None
        else:
            wv0 = buchdahl.get_wv('d')
            gname_decode, rndx, nu1, nu2 = gdata
            g = buchdahl.Buchdahl(wv0, rndx, (nu1, nu2), mat=gname, cat=gcat)
            return g

    def catalog_name(self):
        return 'Robb1983'

    def get_glass_names(self, gcat=None):
        """ returns a list of glass names """
        if gcat is not None:
            return self.glass_data[gcat][0]
        else:  # return all of the catalogs' names
            gnames = [gn[0] for gn in self.glass_data.values()]
            return list(itertools.chain.from_iterable(gnames))

    def glass_map_data(self, wvl='d', **kwargs):
        """ return index and dispersion data for all glasses in the catalog

        Args:
            wvl (str): the central wavelength for the data, either 'd' or 'e'

        Returns:
            index, V-number, partial dispersion, Buchdahl coefficients, and
            glass names
        """
        return self.get_glass_map_arrays(wvl, 'F', 'C', **kwargs)

    def get_glass_map_arrays(self, d_str, F_str, C_str, **kwargs):
        """ return index and dispersion data arrays for input spectral range

        Args:
            nd_str (str): central wavelength string
            nf_str (str): blue end wavelength string
            nc_str (str): red end wavelength string

        Returns:
            index, V-number, partial dispersion, Buchdahl coefficients, and
            glass names
        """
        wv_0 = buchdahl.get_wv('d')
        wv_d = buchdahl.get_wv(d_str)
        wv_F = buchdahl.get_wv(F_str)
        wv_C = buchdahl.get_wv(C_str)

        om_d = buchdahl.omega(wv_d - wv_0)
        om_F = buchdahl.omega(wv_F - wv_0)
        om_C = buchdahl.omega(wv_C - wv_0)

        gnames, gdata = self.glass_data[kwargs['cat_name']]

        omm_d = np.array([om_d, om_d**2])
        nd = np.matmul(gdata[:, 1:], omm_d) + gdata[:, 0]

        omm_F = np.array([om_F, om_F**2])
        nF = np.matmul(gdata[:, 1:], omm_F) + gdata[:, 0]

        omm_C = np.array([om_C, om_C**2])
        nC = np.matmul(gdata[:, 1:], omm_C) + gdata[:, 0]

        vd, PCd = calc_glass_constants(nd, nF, nC)

        coefs = np.array(gdata[:, 1:3])
        ctype = kwargs.get('ctype', None)
        if ctype == "disp_coefs":
            coefs[:, 0] /= (nd - 1)
            coefs[:, 1] /= (nd - 1)

        return nd, vd, PCd, coefs[:, 0], coefs[:, 1], gnames


_robb1983_list = CaselessDictionary({
        'Robb1983.Schott': Robb1983Catalog(),
        'Robb1983.Ohara': Robb1983Catalog(),
        'Robb1983.Hoya': Robb1983Catalog(),
        'Robb1983.Corning-France': Robb1983Catalog(),
        'Robb1983.Chance': Robb1983Catalog(),
        })



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2019 Michael J. Hayford
"""Support for using wavelength values and spectral line designations interchangably.

.. codeauthor: Michael J. Hayford
"""
import numpy as np

spectral_lines = [[2325.42, '', 'infrared mercury line', 'Hg'],
                  [1970.09, '', 'infrared mercury line', 'Hg'],
                  [1529.582, '', 'infrared mercury line', 'Hg'],
                  [1060.0, '', 'neodymium glass laser', 'Nd'],
                  [1013.98, 't', 'infrared mercury line', 'Hg'],
                  [852.11, 's', 'infrared cesium line', 'Cs'],
                  [706.5188, 'r', 'red helium line', 'He'],
                  [656.2725, 'C', 'red hydrogen line', 'H'],
                  [643.8469, "C'", 'red cadmium line', 'Cd'],
                  [632.8, '', 'helium-neon-gas-laser', 'He-Ne'],
                  [589.2938, 'D', 'center of double sodium line', 'Na'],
                  [587.5618, 'd', 'yellow helium line', 'He'],
                  [546.074, 'e', 'green mercury line', 'Hg'],
                  [486.1327, 'F', 'blue hydrogen line', 'H'],
                  [479.9914, "F'", 'blue cadmium line', 'Cd'],
                  [435.8343, 'g', 'blue mercury line', 'Hg'],
                  [404.6561, 'h', 'violet mercury line', 'Hg'],
                  [365.014, 'i', 'ultraviolet mercury line', 'Hg'],
                  [334.1478, '', 'ultraviolet mercury line', 'Hg'],
                  [312.5663, '', 'ultraviolet mercury line', 'Hg'],
                  [296.7278, '', 'ultraviolet mercury line', 'Hg'],
                  [280.4, '', 'ultraviolet mercury line', 'Hg'],
                  [248.3, '', 'ultraviolet mercury line', 'Hg']]


spectra = {'Nd': 1060.0,
           't': 1013.98,
           's': 852.11,
           'r': 706.5188,
           'C': 656.2725,
           "C'": 643.8469,
           'He-Ne': 632.8,
           'D': 589.2938,
           'd': 587.5618,
           'e': 546.074,
           'F': 486.1327,
           "F'": 479.9914,
           'g': 435.8343,
           'h': 404.6561,
           'i': 365.014}
""" dict:

       - keys: spectral line labels
       - values: wavelengths in nm
"""


spectra_uc = {key.upper(): val for key, val in spectra.items()}


def get_wavelength(wvl):
    """Return wvl in nm, where wvl can be a spectral line

    Example::

        In [1]: from rayoptics.util.spectral_lines import *

        In [2]: wl_e = get_wavelength('e'); wl_e
        Out[2]: 546.074

        In [3]: wl_HeNe = get_wavelength('He-Ne'); wl_HeNe
        Out[3]: 632.8

        In [4]: wl_550 = get_wavelength(550); wl_550
        Out[4]: 550.0

        In [5]: wl_fl = get_wavelength(555.0); wl_fl
        Out[5]: 555.0

        In [6]: wl_f = get_wavelength('f'); wl_f
        Out[6]: 486.1327

    Args:
        wvl: either the wavelength in nm or a string with a spectral line
             identifier. Case insensitive.

    Returns:
        float: the wavelength in nm

    Raises:
        KeyError: if ``wvl`` is not in the spectra dictionary
    """
    if isinstance(wvl, np.float):
        return wvl
    elif isinstance(wvl, (int, np.integer)):
        return float(wvl)
    else:
        return spectra_uc[wvl.upper()]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2018 Michael J. Hayford
""" Interfaces for commercial glass catalogs

    The glassfactory module is intended to be the primary method by which glass
    instances are created. The :func:`create_glass` is the public factory
    function for this purpose. The public function :func:`get_glass_catalog`
    returns the glass catalog instance corresponding to the input string.

.. codeauthor: Michael J. Hayford
"""
import logging

from . import cdgm as c
from . import hikari as hi
from . import hoya as h
from . import ohara as o
from . import schott as s
from . import sumita as su

from . import glass as cat_glass
from . import glasserror as ge

from .caselessDictionary import CaselessDictionary

_catalog_list = CaselessDictionary({
    'CDGM': c.CDGMCatalog(),
    'Hikari': hi.HikariCatalog(),
    'Hoya': h.HoyaCatalog(),
    'Ohara': o.OharaCatalog(),
    'Schott': s.SchottCatalog(),
    'Sumita': su.SumitaCatalog(),
    })

CDGM, Hikari, Hoya, Ohara, Schott, Sumita = range(6)
_cat_names = ["CDGM", "Hikari", "Hoya", "Ohara", "Schott", "Sumita"]
_cat_names_uc = [cat.upper() for cat in _cat_names]


def create_glass(*name_catalog):
    """ Factory function returning a catalog glass instance.

    The input argument list can take several forms:

        - 1 string argument in the form 'glass_name,catalog_name'
        - 2 arguments. The first is a string glass name. The second is a
          string or list of strings of catalog names.

    Arguments:
        *name_catalog: tuple of 1 or 2 input items

    Raises:
        GlassCatalogNotFoundError: if catalog isn't found
        GlassNotFoundError: if name isn't in the specified catalog

    """
    def _create_glass(name, catalog):
        gn_decode = cat_glass.decode_glass_name(name)
        if catalog in _catalog_list:
            try:
                # Lookup the decoded glass name., This avoids some problems
                # with how design programs not exactly matching the
                # manufacturer's names.
                gn, gc = _catalog_list[catalog].glass_lookup[gn_decode]
            except KeyError:
                raise ge.GlassNotFoundError(catalog, name)
            else:
                return _catalog_list[catalog].create_glass(gn, gc)
        elif "Robb1983" in catalog:
            return cat_glass.Robb1983Catalog().create_glass(name, catalog)
        else:
            logging.info('glass catalog %s not found', catalog)
            raise ge.GlassCatalogNotFoundError(catalog)

    if len(name_catalog) == 2:
        name, catalog = name_catalog
    else:
        name, catalog = name_catalog[0].split(',')

    if isinstance(catalog, str):
        return _create_glass(name, catalog)

    else:  # treat catalog as a list
        for cat in catalog:
            try:
                glass = _create_glass(name, cat)
            except ge.GlassError:
                continue
            else:
                return glass
        logging.info('glass %s not found in %s', name, catalog)
        raise ge.GlassNotFoundError(catalog, name)


def get_glass_catalog(catalog):
    """ Function returning a glass catalog instance.

    Arguments:
        catalog: name of supported catalog (CDGM, Hoya, Ohara, Schott)

    Raises:
        GlassCatalogNotFoundError: if catalog isn't found
    """
    if catalog in _catalog_list:
        return _catalog_list[catalog]
    else:
        logging.info('glass catalog %s not found', catalog)
        raise ge.GlassCatalogNotFoundError(catalog)
        return None



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
""" unit test for Hikari glass catalog

.. Created on Wed Aug 26 15:12:40 2020

.. codeauthor: Michael J. Hayford
"""

import unittest
import opticalglass.hikari as hi


class HikariTestCase(unittest.TestCase):
    catalog = hi.HikariCatalog()

    def compare_indices(self, glass, tol=5e-6):
        nC = glass.rindex(656.27)
        nd = glass.rindex(587.56)
        ne = glass.rindex(546.07)
        nF = glass.rindex(486.13)
        ng = glass.rindex(435.84)
        nh = glass.rindex(404.66)
        nI = glass.rindex(365.01)
        nline = self.catalog.nline_str
        indxC = glass.glass_item(nline['C'])
        indxd = glass.glass_item(nline['d'])
        indxe = glass.glass_item(nline['e'])
        indxF = glass.glass_item(nline['F'])
        indxg = glass.glass_item(nline['g'])
        indxh = glass.glass_item(nline['h'])
        indxI = glass.glass_item(nline['i'])
        self.assertAlmostEqual(nC, indxC, delta=tol)
        self.assertAlmostEqual(nd, indxd, delta=tol)
        self.assertAlmostEqual(ne, indxe, delta=tol)
        self.assertAlmostEqual(nF, indxF, delta=tol)
        self.assertAlmostEqual(ng, indxg, delta=tol)
        self.assertAlmostEqual(nh, indxh, delta=tol)
        self.assertAlmostEqual(nI, indxI, delta=tol)

    def test_hikari_catalog_glass_index(self):
        fk5 = self.catalog.glass_index('J-FK5')  # first in list
        self.assertEqual(fk5, 0)
        bk7a = self.catalog.glass_index('J-BK7A')
        self.assertEqual(bk7a, 9)
        sk16 = self.catalog.glass_index('J-SK16')
        self.assertEqual(sk16, 39)
        lasf02 = self.catalog.glass_index('J-LASF02')
        self.assertEqual(lasf02, 104)
        lasfh24hs = self.catalog.glass_index('J-LASFH24HS')  # last in list
        self.assertEqual(lasfh24hs, 131)

    def test_hikari_catalog_data_index(self):
        nd = self.catalog.data_index(self.catalog.nline_str['d'])
        self.assertEqual(nd, 16)
        vd = self.catalog.data_index('νd')
        self.assertEqual(vd, 24)
        A5 = self.catalog.data_index('A5/λ^6')
        self.assertEqual(A5, 60)
        glasscode = self.catalog.data_index('Code(d)')
        self.assertEqual(glasscode, 1)
        date = self.catalog.data_index('550nm')
        self.assertEqual(date, 115)

    def test_hikari_glass_bk7(self):
        glass = hi.HikariGlass('J-BK7A')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'J-BK7A')
        self.compare_indices(glass, tol=1e-6)

    def test_hikari_glass_sk16(self):
        glass = hi.HikariGlass('J-SK16')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'J-SK16')
        self.compare_indices(glass, tol=1.05e-6)

    def test_hikari_glass_lasf02(self):
        glass = hi.HikariGlass('J-LASF02')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'J-LASF02')
        self.compare_indices(glass, tol=2.25e-6)


if __name__ == '__main__':
    unittest.main(verbosity=2)






#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Support for the Hoya Glass catalog

.. codeauthor: Michael J. Hayford
"""

import numpy as np
from .util import Singleton

from . import glass


class HoyaCatalog(glass.GlassCatalogXLSX, metaclass=Singleton):
    #    data_header = 1
    #    data_start = 4
    #    num_glasses = 194
    #    name_col_offset = 2
    #    coef_col_offset = 28
    #    index_col_offset = 10
    nline_str = {'t': 'nt',
                 's': 'ns',
                 'r': 'nr',
                 'C': 'nC',
                 "C'": "nC'",
                 'D': 'nD',
                 'd': 'nd',
                 'e': 'ne',
                 'F': 'nF',
                 "F'": "nF'",
                 'g': 'ng',
                 'h': 'nh',
                 'i': 'ni'}

    def __init__(self, fname='HOYA.xlsx'):
        super().__init__('Hoya', fname, 'Glass\u3000Type', 'A0', 'n1529.6',
                         num_coefs=12, data_header_offset=1,
                         transmission_offset=464, num_wvls=44,
                         transmission_header_offset=2)

    def glass_coefs(self, gindex):
        c = super().glass_coefs(gindex)
        coefs = [x*10**y for x, y in zip(c[::2], c[1::2])]
        return coefs

    def create_glass(self, gname, gcat):
        return HoyaGlass(gname)


class HoyaGlass(glass.Glass):
    catalog = HoyaCatalog()

    def __init__(self, gname, catalog=None):
        if catalog is not None:
            self.catalog = catalog
        super().__init__(gname)

    def glass_code(self):
        return super().glass_code('nd', 'νd')

    def calc_rindex(self, wv_nm):
        wv = 0.001*wv_nm
        wv2 = wv*wv
        coefs = self.coefs
        n2 = coefs[0] + coefs[1]*wv2
        wvm2 = 1/wv2
        n2 = n2 + wvm2*(coefs[2] + wvm2*(coefs[3]
                        + wvm2*(coefs[4] + wvm2*coefs[5])))
        return np.sqrt(n2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" unit test for CDGM optical glass catalog

.. codeauthor: Michael J. Hayford
"""

import unittest
import opticalglass.cdgm as c


class CDGMTestCase(unittest.TestCase):
    catalog = c.CDGMCatalog()

    def compare_indices(self, glass, tol=5e-6):
        nC = glass.rindex(656.27)
        nd = glass.rindex(587.56)
        ne = glass.rindex(546.07)
        nF = glass.rindex(486.13)
        ng = glass.rindex(435.84)
        nh = glass.rindex(404.66)
        nI = glass.rindex(365.01)
        nline = self.catalog.nline_str
        indxC = glass.glass_item(nline['C'])
        indxd = glass.glass_item(nline['d'])
        indxe = glass.glass_item(nline['e'])
        indxF = glass.glass_item(nline['F'])
        indxg = glass.glass_item(nline['g'])
        indxh = glass.glass_item(nline['h'])
        indxI = glass.glass_item(nline['i'])
        self.assertAlmostEqual(nC, indxC, delta=tol)
        self.assertAlmostEqual(nd, indxd, delta=tol)
        self.assertAlmostEqual(ne, indxe, delta=tol)
        self.assertAlmostEqual(nF, indxF, delta=tol)
        self.assertAlmostEqual(ng, indxg, delta=tol)
        self.assertAlmostEqual(nh, indxh, delta=tol)
        self.assertAlmostEqual(nI, indxI, delta=tol)

    def test_cdgm_catalog_glass_index(self):
        g1 = self.catalog.glass_index('H-FK61')  # first in list
        self.assertEqual(g1, 0)
        g2 = self.catalog.glass_index('H-QK3L')
        self.assertEqual(g2, 4)
        g3 = self.catalog.glass_index('F4')
        self.assertEqual(g3, 97)
        g4 = self.catalog.glass_index('H-K9L')
        self.assertEqual(g4, 13)
        g5 = self.catalog.glass_index('D-LaF50')
        self.assertEqual(g5, 229)
        g6 = self.catalog.glass_index('D-ZLaF85L')  # last in list
        self.assertEqual(g6, 239)

    def test_cdgm_catalog_data_index(self):
        nd = self.catalog.data_index('nd')
        self.assertEqual(nd, 9)
        vd = self.catalog.data_index('υd')
        self.assertEqual(vd, 17)
        A0 = self.catalog.data_index('A0')
        self.assertEqual(A0, 21)
        nt = self.catalog.data_index('nt')
        self.assertEqual(nt, 2)

    def test_cdgm_glass_hfk61(self):
        glass = c.CDGMGlass('H-FK61')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'H-FK61')
        self.compare_indices(glass, tol=6e-6)

    def test_cdgm_glass_f4(self):
        glass = c.CDGMGlass('F4')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'F4')
        self.compare_indices(glass, tol=1e-5)

    def test_cdgm_glass_hk9l(self):
        glass = c.CDGMGlass('H-K9L')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'H-K9L')
        self.compare_indices(glass)


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" unit test for GlassCatalogSpreadsheet transmission_data function

.. codeauthor: Michael J. Hayford
"""

import unittest
import opticalglass.glassfactory as gf


class TransmissionTestCase(unittest.TestCase):

    def compare_transmission(self, glass, wl1, t1, wl2, t2, num_t, tol=5e-4):
        t_data = glass.transmission_data()
        self.assertEqual(wl1, t_data[0][0])
        self.assertAlmostEqual(t1, t_data[0][1], delta=tol)
        self.assertEqual(wl2, t_data[-1][0])
        self.assertAlmostEqual(t2, t_data[-1][1], delta=tol)
        self.assertEqual(num_t, len(t_data))

    def test_cdgm_glass_hlak50a(self):
        glass = gf.create_glass('H-LaK50A', 'CDGM')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'H-LaK50A')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 2400.0, 0.668, 290.0, 0.028, 34)

    def test_cdgm_glass_hlak51a(self):
        glass = gf.create_glass('H-LaK51A', 'CDGM')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'H-LaK51A')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 2400.0, 0.633, 290.0, 0.218, 34)

    def test_hikari_glass_jlak8(self):
        glass = gf.create_glass('J-LAK8', 'Hikari')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'J-LAK8')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 280.0, 0.01, 2400.0, 0.63, 33)

    def test_hikari_glass_jlasf010(self):
        glass = gf.create_glass('J-LASF010', 'Hikari')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'J-LASF010')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 350.0, 0.14, 2400.0, 0.72, 26)

    def test_hoya_glass_pcd51(self):
        glass = gf.create_glass('PCD51', 'Hoya')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'PCD51')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 2500.0, 0.8, 280.0, 0.145, 44)

    def test_hoya_glass_fd140(self):
        glass = gf.create_glass('FD140', 'Hoya')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'FD140')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 1550.0, 0.995, 280.0, 0, 38)

    def test_ohara_glass_slah51(self):
        glass = gf.create_glass('S-LAH51', 'Ohara')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'S-LAH51')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 330.0, 0.01, 2400.0, 0.68, 27)

    def test_ohara_glass_slah97(self):
        glass = gf.create_glass('S-LAH97', 'Ohara')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'S-LAH97')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 280.0, 0.28, 2400.0, 0.61, 32)

    def test_schott_glass_lf5(self):
        glass = gf.create_glass('LF5', 'Schott')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'LF5')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 2325.0, 0.847, 310.0, 0.04, 23)

    def test_schott_glass_nbk7(self):
        glass = gf.create_glass('N-BK7', 'Schott')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'N-BK7')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 2500.0, 0.665, 290.0, 0.063, 26)

    def test_schott_glass_nfk58(self):
        glass = gf.create_glass('N-FK58', 'Schott')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'N-FK58')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 2500.0, 0.997, 250.0, 0.09, 30)

    def test_sumita_glass_ksk16(self):
        glass = gf.create_glass('K-SK16', 'Sumita')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'K-SK16')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 300.0, 0.014, 2000.0, 0.937, 24)

    def test_sumita_glass_kbk7(self):
        glass = gf.create_glass('K-BK7', 'Sumita')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'K-BK7')
        # t_data = glass.transmission_data()
        # print(glass.gname, glass.gindex, t_data[0][0], t_data[0][1],
        #       t_data[-1][0], t_data[-1][1], len(t_data))
        self.compare_transmission(glass, 270.0, 0.02, 2000.0, 0.959, 27)


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2017 Michael J. Hayford
""" Support for the Schott Glass catalog

.. codeauthor: Michael J. Hayford
"""
import logging
from .util import Singleton

import numpy as np

from . import glass


class SchottCatalog(glass.GlassCatalogXLS, metaclass=Singleton):
    #    data_header = 3
    #    data_start = 4
    #    num_glasses = 123
    #    name_col_offset = 0
    #    coef_col_offset = 6
    #    index_col_offset = 117
    nline_str = {'t': '  nt',
                 's': '  ns',
                 'r': '  nr',
                 'C': '  nC',
                 "C'": "  nC'",
                 'D': '  nD',
                 'd': '  nd',
                 'e': '  ne',
                 'F': '  nF',
                 "F'": "  nF'",
                 'g': '  ng',
                 'h': '  nh',
                 'i': '  ni'}

    def __init__(self, fname='SCHOTT.xls'):
        super().__init__('Schott', fname, 'Glass', 'B1', '  n2325.4',
                         transmission_offset=67, num_wvls=30)

    def create_glass(self, gname, gcat):
        return SchottGlass(gname)

    def get_transmission_wvl(self, header_str):
        """Returns the wavelength value from the transmission data header string."""
        return float(header_str[len('TAUI10/'):])


class SchottGlass(glass.Glass):
    catalog = SchottCatalog()

    def __init__(self, gname, catalog=None):
        if catalog is not None:
            self.catalog = catalog
        super().__init__(gname)

    def glass_code(self):
        return super().glass_code('nd', 'vd')

    def calc_rindex(self, wv_nm):
        wv = 0.001*wv_nm
        wv2 = wv*wv
        coefs = self.coefs
        n2 = 1. + coefs[0]*wv2/(wv2 - coefs[3])
        n2 = n2 + coefs[1]*wv2/(wv2 - coefs[4])
        n2 = n2 + coefs[2]*wv2/(wv2 - coefs[5])
        return np.sqrt(n2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
""" unit test for Sumita optical glass catalog

.. Created on Wed Aug 26 14:46:13 2020

.. codeauthor: Michael J. Hayford
"""

import unittest
import opticalglass.sumita as su


class SumitaTestCase(unittest.TestCase):
    catalog = su.SumitaCatalog()

    def compare_indices(self, glass, tol=5e-6):
        nC = glass.rindex(656.27)
        nd = glass.rindex(587.56)
        ne = glass.rindex(546.07)
        nF = glass.rindex(486.13)
        ng = glass.rindex(435.84)
        nh = glass.rindex(404.66)
        nI = glass.rindex(365.01)
        nline = self.catalog.nline_str
        indxC = glass.glass_item(nline['C'])
        indxd = glass.glass_item(nline['d'])
        indxe = glass.glass_item(nline['e'])
        indxF = glass.glass_item(nline['F'])
        indxg = glass.glass_item(nline['g'])
        indxh = glass.glass_item(nline['h'])
        indxI = glass.glass_item(nline['i'])
        self.assertAlmostEqual(nC, indxC, delta=tol)
        self.assertAlmostEqual(nd, indxd, delta=tol)
        self.assertAlmostEqual(ne, indxe, delta=tol)
        self.assertAlmostEqual(nF, indxF, delta=tol)
        self.assertAlmostEqual(ng, indxg, delta=tol)
        self.assertAlmostEqual(nh, indxh, delta=tol)
        self.assertAlmostEqual(nI, indxI, delta=tol)

    def test_ohara_catalog_glass_index(self):
        cafk95 = self.catalog.glass_index('K-CaFK95')  # first in list
        self.assertEqual(cafk95, 1)
        pbk40 = self.catalog.glass_index('K-PBK40')
        self.assertEqual(pbk40, 7)
        sk16 = self.catalog.glass_index('K-SK16')
        self.assertEqual(sk16, 64)
        laskn1 = self.catalog.glass_index('K-LaSKn1')
        self.assertEqual(laskn1, 101)
        fir100uv = self.catalog.glass_index('K-FIR100UV')  # last in list
        self.assertEqual(fir100uv, 134)

    def test_sumita_catalog_data_index(self):
        nd = self.catalog.data_index('nd')
        self.assertEqual(nd, 16)
        vd = self.catalog.data_index('vd')
        self.assertEqual(vd, 4)
        B1 = self.catalog.data_index('A4')
        self.assertEqual(B1, 52)
        glasscode = self.catalog.data_index('GTYPE')
        self.assertEqual(glasscode, 2)
        date = self.catalog.data_index('T1_550')
        self.assertEqual(date, 97)

    def test_sumita_glass_pbk40(self):
        glass = su.SumitaGlass('K-PBK40')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'K-PBK40')
        self.compare_indices(glass, tol=1.1e-5)

    def test_sumita_glass_sk16(self):
        glass = su.SumitaGlass('K-SK16')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'K-SK16')
        self.compare_indices(glass, tol=2.5e-6)

    def test_sumita_glass_laskn1(self):
        glass = su.SumitaGlass('K-LaSKn1')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'K-LaSKn1')
        self.compare_indices(glass, tol=7.5e-6)


if __name__ == '__main__':
    unittest.main(verbosity=2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
"""

.. Created on Thu Dec 24 17:17:36 2020

.. codeauthor: Michael J. Hayford
"""
import warnings

from openpyxl import load_workbook
from openpyxl import worksheet

from opticalglass import ohara
import opticalglass.glass as catgl
from opticalglass import glassfactory as gfact

NBK7 = gfact.create_glass('N-BK7', 'Schott')
tdat_nbk7 = NBK7.transmission_data()


LaK10 = gfact.create_glass('K-LaK10', 'Sumita')

sumita_cat = LaK10.catalog

tdat = LaK10.transmission_data()


fname = 'OHARA.xlsx'
with warnings.catch_warnings():
    warnings.filterwarnings("ignore", category=UserWarning)
    wb = load_workbook(catgl.get_filepath(fname))
ws = wb.active

glass_str = 'Glass '
data_header_offset = 0
glass_name_offset = 1
for i, r in enumerate(ws.iter_rows(values_only=True), start=1):
    try:
        name_col_offset = r.index(glass_str) + 1  # one based
    except ValueError:
        pass
    else:
        glass_header = i
        data_header = i
        # the data headers may be offset from the Glass header row
        data_header += data_header_offset
        break

cols = ws.iter_cols(min_col=name_col_offset,
                    max_col=name_col_offset,
                    values_only=True)
gnames = list(next(cols))
for j, gname in enumerate(gnames[data_header:], start=data_header+1):
    if len(gname) > 0:
        data_start = j
        break

while len(gnames) > 0:
    if gnames[-1] is None:
        gnames.pop()
    elif len(gnames[-1]) == 0:
        gnames.pop()
    else:
        break

num_glasses = len(gnames[data_header:])

for j, c in enumerate(ws.iter_cols(min_col=name_col_offset,
                                   max_col=name_col_offset,
                                   values_only=True), start=1):
    if len(c) > 0:
        data_start = j
        break

cols = ws.iter_cols(min_col=name_col_offset, max_col=name_col_offset,
                    min_row=data_start, max_row=ws.max_row, values_only=True)
gnames = list(cols)
while len(gnames) > 0 and len(gnames[-1]) == 0:
    gnames.pop()

# col = ws.iter_cols(min_col=name_col_offset, max_col=name_col_offset,
#                    min_row=1, max_row=num_glasses, values_only=True)
ws.title
print(wb.sheetnames)
ws.cell(2,3)
ws.cell(3,3)
ws.cell(3,5)
c35=ws.cell(3,5)
c35.value
ws.max_row
#len(tuple(ws.rows))

# offsets for Ohara
#data_header = 1
#data_start = 2+1
num_glasses = 134
#name_col_offset = 1+1
#coef_col_offset = 60
#index_col_offset = 4

ws.dimensions
ws.max_row
ws.max_column

cols = ws.iter_cols(min_col=name_col_offset, max_col=name_col_offset,
                    min_row=data_start, max_row=num_glasses, values_only=True)

name_col = next(cols)
# for i, n in enumerate(name_col):
#     print(i, n)

ohara_cat = ohara.OharaCatalog()
ohara_gnames = ohara_cat.get_glass_names()
for i, n in enumerate(ohara_gnames):
    print(i, n)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
"""Buchdahl chromatic coordinate modeling and support

.. Created on Thu Aug 20 21:27:36 2020

.. codeauthor: Michael J. Hayford
"""

import numpy as np

from .spectral_lines import get_wavelength


def get_wv(wavelength):
    """Return the wavelength in micrometers."""
    return get_wavelength(wavelength)*1e-3


def omega(delta_lambda):
    """Calculate the Buchdahl chromatic coordinate."""
    return delta_lambda/(1 + 2.5*delta_lambda)


def omega2wvl(om):
    return om/(1 - 2.5*om)


def model_from_glasses(gla1, gla2):
    """Create a model from the slope between two glasses."""
    bhdl_gla1 = Buchdahl1(gla1)
    bhdl_gla2 = Buchdahl1(gla2)
    # get the Buchdahl quadratic coefficients for the 2 input glasses
    v1_gla1, v2_gla1 = bhdl_gla1.coefs[0], bhdl_gla1.coefs[1]
    v1_gla2, v2_gla2 = bhdl_gla2.coefs[0], bhdl_gla2.coefs[1]
    # calculate the slope and v1 intercept of the line between the 2 glasses
    m = (v1_gla1 - v1_gla2)/(v2_gla1 - v2_gla2)
    b = v1_gla1 - m*v2_gla1
    return b, m


class Buchdahl:
    """Quadratic Buchdahl refractive index model.

    .. math::

        N(\omega) = {N_0} + \\nu_1\omega + \\nu_2\omega^2

    - :math:`\omega` is the Buchdahl chromatic coordinate for the input wavelength
    - :math:`\\nu_1, \\nu_2` are the linear and quadratic coefficients of the model
    - :math:`{N_0}` is the refractive index at the central wavelength of the fit

    The Buchdahl chromatic coordinate :math:`\omega` is defined as:

    .. math::

        \omega(\lambda) = \\frac{\lambda - \lambda_0}{1 + 5/2(\lambda - \lambda_0)}

    """

    def __init__(self, wv0, rind0, coefs, mat='', cat=''):
        """
        Parameters
        ----------
        wv0 : float
            central wavelength in **micrometers**.
        rind0 : float
            refractive index at the central wavelength.
        coefs : (float, float)
            the linear and quadratic coefficients of the model.
        mat : str
            a string label returned from the name() fct.
        cat : str
            a string label returned from the catalog_name() fct.
        """
        self.wv0 = wv0
        self.rind0 = rind0
        self.coefs = coefs
        self.label = mat
        self._catalog_name = cat

    def name(self):
        return self.label

    def catalog_name(self):
        return self._catalog_name

    def glass_code(self):
        nd = self.rindex('d')
        nF = self.rindex('F')
        nC = self.rindex('C')
        vd = (nd - 1)/(nF - nC)
        return str(1000*round((nd - 1), 3) + round(vd/100, 3))

    def rindex(self, wvl):
        """Returns the refractive index from the quadratic model at wvl."""
        return self.calc_rindex(get_wavelength(wvl))

    def calc_rindex(self, wv_nm):
        om = omega((wv_nm*1e-3) - self.wv0)
        return self.rind0 + self.coefs[0]*om + self.coefs[1]*om**2


class Buchdahl1(Buchdahl):
    """Quadratic refractive index model for a real glass, *medium*. """

    def __init__(self, medium, wlns=('d', 'F', 'C'), **kwargs):
        rindx = [medium.rindex(wlns[0]),
                 medium.rindex(wlns[1]),
                 medium.rindex(wlns[2])]

        wv0 = get_wv(wlns[0])
        omF = omega(get_wv(wlns[1]) - wv0)
        omC = omega(get_wv(wlns[2]) - wv0)
        self.om = omF, omC

        coefs = self.update(rindx)
        super().__init__(wv0, rindx[0], coefs, **kwargs)

    def update(self, rindx):
        omF, omC = self.om
        a = np.array([[omF, omF**2], [omC, omC**2]])
        b = np.array([rindx[1]-rindx[0], rindx[2]-rindx[0]])
        coefs = np.linalg.solve(a, b)
        return coefs


class Buchdahl2(Buchdahl):
    """Quadratic refractive index model for a 6-digit glass specification. """

    b = -0.064667
    m = -1.604048

    def __init__(self, nd, vd, model=None, wlns=('d', 'F', 'C'), **kwargs):
        if model is not None:
            self.b, self.m = model
        wv0 = get_wv(wlns[0])
        omF = omega(get_wv(wlns[1]) - wv0)
        omC = omega(get_wv(wlns[2]) - wv0)
        delta_om = omF - omC
        delta_om2 = omF**2 - omC**2
        self.om = omF, omC, delta_om, delta_om2
        rind0, coefs = self.update(nd, vd)
        super().__init__(wv0, rind0, coefs, **kwargs)

    def update(self, nd, vd):
        omF, omC, delta_om, delta_om2 = self.om
        dFC = (nd - 1.0)/vd
        v2 = (dFC - self.b*delta_om)/(self.m*delta_om - delta_om2)
        v1 = self.b + self.m*v2
        return nd, np.array([v1, v2])

    def update_model(self, nd, vd):
        rind0, coefs = self.update(nd, vd)
        self.rind0 = rind0
        self.coefs = coefs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright © 2020 Michael J. Hayford
"""Support for the Hikari Glass catalog

.. Created on Tue Aug 25 22:55:44 2020

.. codeauthor: Michael J. Hayford
"""

import logging
from .util import Singleton

import numpy as np

from . import glass


class HikariCatalog(glass.GlassCatalogXLS, metaclass=Singleton):
    #    data_header = 0
    #    data_start = 2
    #    num_glasses = 240
    #    name_col_offset = 0
    #    coef_col_offset = 21
    #    index_col_offset = 2
    nline_str = {'t': 't\n1.01398',
                 's': 's\n0.85211',
                 'r': 'r\n0.706519',
                 'C': 'C\n0.656273',
                 "C'": "C'\n0.643847",
                 'D': 'D\n0.589294',
                 'd': 'd\n0.587562',
                 'e': 'e\n0.546074',
                 'F': 'F\n0.486133',
                 "F'": "nF'\n0.479992",
                 'g': 'g\n0.435835',
                 'h': 'h\n0.404656',
                 'i': 'i\n0.365015'}

    def __init__(self, fname='HIKARI.xls'):
        super().__init__('Hikari', fname, 'Glass type', 'A0', 2.05809,
                         data_header_offset=1, glass_name_offset=2,
                         num_coefs=9,
                         transmission_offset=96, num_wvls=33)

    def create_glass(self, gname, gcat):
        return HikariGlass(gname)

    def get_transmission_wvl(self, header_str):
        """Returns the wavelength value from the transmission data header string."""
        return float(header_str[:-len('nm')])


class HikariGlass(glass.Glass):
    catalog = HikariCatalog()

    def __init__(self, gname, catalog=None):
        if catalog is not None:
            self.catalog = catalog
        super().__init__(gname)

    def glass_code(self):
        return super().glass_code('d\n0.587562', 'νd')

    def calc_rindex(self, wv_nm):
        wv = 0.001*wv_nm
        wv2 = wv*wv
        coefs = self.coefs
        n2 = coefs[0] + wv2*(coefs[1] + wv2*coefs[2])
        wvm2 = 1/wv2
        n2 = n2 + wvm2*(coefs[3] +
                        wvm2*(coefs[4] +
                              wvm2*(coefs[5] +
                                    wvm2*(coefs[6] +
                                          wvm2*(coefs[7] +
                                                wvm2*coefs[8])))))
        return np.sqrt(n2)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="opticalglass",
    version="0.7.3",
    author="Michael J Hayford",
    author_email="mjhoptics@gmail.com",
    description="Tools for reading commercial optical glass catalogs",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    license="BSD-3-Clause",
    url="https://github.com/mjhoptics/opticalglass",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    keywords=['glass', 'optical glass', 'refractive index', 'optics',
              'glass catalog'],
    install_requires=[
        "xlrd>=2.0.1",
        "openpyxl>=3.0.5",
        "numpy>=1.17.1",
        "scipy>=1.3.1",
        "matplotlib>=3.1.1",
        "pyqt5<5.13"
        ],
    package_data={
        '': ['data/*.xlsx', 'data/*.xls', 'data/*.txt']
        },
    entry_points={
        'gui_scripts': [
            'glassmap = opticalglass.glassmapviewer:main',
        ],
    },
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" unit test for GlassFactory create_glass function

.. codeauthor: Michael J. Hayford
"""

import unittest
import opticalglass.glassfactory as gf


class GlassFactoryTestCase(unittest.TestCase):

    def compare_indices(self, glass, tol=5e-6):
        nC = glass.rindex(656.27)
        nd = glass.rindex(587.56)
        ne = glass.rindex(546.07)
        nF = glass.rindex(486.13)
        ng = glass.rindex(435.84)
        nh = glass.rindex(404.66)
        nI = glass.rindex(365.01)
        nline = glass.catalog.nline_str
        indxC = glass.glass_item(nline['C'])
        indxd = glass.glass_item(nline['d'])
        indxe = glass.glass_item(nline['e'])
        indxF = glass.glass_item(nline['F'])
        indxg = glass.glass_item(nline['g'])
        indxh = glass.glass_item(nline['h'])
        indxI = glass.glass_item(nline['i'])
        self.assertAlmostEqual(nC, indxC, delta=tol)
        self.assertAlmostEqual(nd, indxd, delta=tol)
        self.assertAlmostEqual(ne, indxe, delta=tol)
        self.assertAlmostEqual(nF, indxF, delta=tol)
        self.assertAlmostEqual(ng, indxg, delta=tol)
        self.assertAlmostEqual(nh, indxh, delta=tol)
        self.assertAlmostEqual(nI, indxI, delta=tol)

    def test_cdgm_glass_hfk61(self):
        glass = gf.create_glass('H-FK61', 'CDGM')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'H-FK61')
        self.compare_indices(glass, tol=6e-6)

    def test_cdgm_glass_f4(self):
        glass = gf.create_glass('F4', 'CDGM')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'F4')
        self.compare_indices(glass, tol=1e-5)

    def test_cdgm_glass_hk9l(self):
        glass = gf.create_glass('H-K9L', 'CDGM')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'H-K9L')
        self.compare_indices(glass)

    def test_hoya_glass_fcd1(self):
        fcd1 = gf.create_glass('FCD1', 'Hoya')
        self.assertIsNotNone(fcd1.gindex)
        self.assertEqual(fcd1.name(), 'FCD1')
        self.compare_indices(fcd1)

    def test_hoya_glass_ef2(self):
        ef2 = gf.create_glass('E-F2', 'Hoya')
        self.assertIsNotNone(ef2.gindex)
        self.assertEqual(ef2.name(), 'E-F2')
        self.compare_indices(ef2)

    def test_hoya_glass_bsc7(self):
        bsc7 = gf.create_glass('BSC7', 'Hoya')
        self.assertIsNotNone(bsc7.gindex)
        self.assertEqual(bsc7.name(), 'BSC7')
        self.compare_indices(bsc7)

    def test_ohara_glass_stim2(self):
        stim2 = gf.create_glass('S-TIM 2', 'Ohara')
        self.assertIsNotNone(stim2.gindex)
        self.assertEqual(stim2.name(), 'S-TIM 2')
        self.compare_indices(stim2, tol=6e-6)

    def test_ohara_glass_sbsl7(self):
        sbsl7 = gf.create_glass('S-BSL 7', 'Ohara')
        self.assertIsNotNone(sbsl7.gindex)
        self.assertEqual(sbsl7.name(), 'S-BSL 7')
        self.compare_indices(sbsl7, tol=6e-6)

    def test_ohara_glass_snbh53v(self):
        glass = gf.create_glass('S-NBH53V', 'Ohara')
        self.assertIsNotNone(glass.gindex)
        self.assertEqual(glass.name(), 'S-NBH53V')
        self.compare_indices(glass)

    def test_schott_glass_f2(self):
        f2 = gf.create_glass('F2', 'Schott')
        self.assertIsNotNone(f2.gindex)
        self.assertEqual(f2.name(), 'F2')
        self.compare_indices(f2)

    def test_schott_glass_nbk7(self):
        nbk7 = gf.create_glass('N-BK7', 'Schott')
        self.assertIsNotNone(nbk7.gindex)
        self.assertEqual(nbk7.name(), 'N-BK7')
        self.compare_indices(nbk7)

    def test_schott_glass_sf6ht(self):
        sf6ht = gf.create_glass('SF6HT', 'Schott')
        self.assertIsNotNone(sf6ht.gindex)
        self.assertEqual(sf6ht.name(), 'SF6HT')
        self.compare_indices(sf6ht)


if __name__ == '__main__':
    unittest.main(verbosity=2)
